/*
*
* Template Numerical Toolkit (TNT)
*
* Mathematical and Computational Sciences Division
* National Institute of Technology,
* Gaithersburg, MD USA
*
*
* This software was developed at the National Institute of Standards and
* Technology (NIST) by employees of the Federal Government in the course
* of their official duties. Pursuant to title 17 Section 105 of the
* United States Code, this software is not subject to copyright protection
* and is in the public domain. NIST assumes no responsibility whatsoever for
* its use by other parties, and makes no guarantees, expressed or implied,
* about its quality, reliability, or any other characteristic.
*
*/



#ifndef TNT_ARRAY2D_H
#define TNT_ARRAY2D_H

#include <cstdlib>
#include <iostream>
#ifdef TNT_BOUNDS_CHECK
#include <assert.h>
#endif

#include "tnt_array1d.h"

namespace TNT
{

template <class T>
class Array2D 
{
  private:

  	Array1D<T> data_;
	Array1D<T*> v_;
	int m_;
    int n_;

  public:

    typedef         T   value_type;
	       Array2D();
	       Array2D(int m, int n);
	       Array2D(int m, int n,  T *a);
	       Array2D(int m, int n, const T &a);
    inline Array2D(const Array2D &A);
	inline operator T**();
	inline operator const T**();
	inline Array2D & operator=(const T &a);
	inline Array2D & operator=(const Array2D &A);
	inline Array2D & ref(const Array2D &A);
	       Array2D copy() const;
		   Array2D & inject( const Array2D & A );
		   Array2D & inject( int iRows, int iCols, const T ** pdA );
	inline T* operator[](int i);
	inline const T* operator[](int i) const;
	inline int dim1() const;
	inline int dim2() const;
     ~Array2D();

	/* extended interface (not part of the standard) */
	inline int ref_count();
	inline int ref_count_data();
	inline int ref_count_dim1();
	Array2D subarray(int i0, int i1, int j0, int j1);

	// JTB extensions
	void SwapRows( int i0, int i1 );		// swap rows i0,i1
	void SwapCols( int i0, int i1 );		// swap columns i0,i1
	void TransposeSquare();
	Array2D<T> Transpose();
	
	T Maximum();
	T Minimum();
	T Sum();
	Array2D<T> & MultiplyAdd( T tMult, T tAdd );
	void SetDiagonal( T d );
};

//----- Adjust all elements : t'= tMult * t + tAdd 
template <class T>
Array2D<T> & Array2D<T>::MultiplyAdd( T tMult, T tAdd ) { 
	for( int i=0; i<m_; ++i ) 
		for( int j=0; j<n_; ++j ) 
			v_[i][j] = v_[i][j]*tMult+tAdd;
	return *this;
	}

//----- swap columns i0 and i1 
template <class T>
void SwapCols( int iRows, int i0, int i1, T ** adA ){ 		
	for( int iRow=0; iRow<iRows; ++iRow ) 
		Swap( adA[iRow]+i0, adA[iRow]+i1 );
	}


//----- Transpose array in place 
template <class T>
void Array2D<T>::TransposeSquare(){
	int iMax = m_<n_?m_:n_;
	T d;
	for( int i=0; i<iMax; ++i ) 
		for( int j=i+1; j<iMax; ++j ) {
			d = v_[i][j];
			v_[i][j] = v_[j][i];
			v_[j][i] = d;
			}
	}

//----- Return Transpose array 
template <class T>
Array2D<T> Array2D<T>::Transpose(){
	Array2D<T> aT( n_, m_ );
	for( int i=0; i<m_; ++i ) 
		for( int j=0; j<n_; ++j ) 
			aT[j][i]=v_[i][j];
	return aT;
	}

//----- Maximum of all elements
template <class T>
T Array2D<T>::Maximum(){
	if( m_==0 || n_==0 ) 
		return 0.0;
	T dMax=v_[0][0];
	
	for( int i=0; i<m_; ++i ) 
		for( int j=0; j<n_; ++j )
			if( v_[i][j] > dMax ) 
				dMax=v_[i][j];
	return dMax;
	}

//----- Minimum of all elements
template <class T>
T Array2D<T>::Minimum(){
	if( m_==0 || n_==0 ) 
		return 0.0;
	T dMin=v_[0][0];
	
	for( int i=0; i<m_; ++i ) 
		for( int j=0; j<n_; ++j )
			if( v_[i][j] < dMin ) 
				dMin=v_[i][j];
	return dMin;
	}

//----- Sum of all elements
template <class T>
T Array2D<T>::Sum(){
	T dsum=0.0;
	for( int i=0; i<m_; ++i ) 
		for( int j=0; j<n_; ++j )
			dsum+=v_[i][j];
	return dSum;
	}

template <class T>
void Array2D<T>::SetDiagonal( T d ) { 
	for( int i=min(m_,n_)-1; i>=0; --i ) 
		v_[i][i]=d;
	}

//----------------------------------------------------------------------


template <class T>
Array2D<T>::Array2D() : data_(), v_(), m_(0), n_(0) {
	} 

template <class T>
Array2D<T>::Array2D(const Array2D<T> &A) : data_(A.data_), v_(A.v_), 
	m_(A.m_), n_(A.n_) {
	}

template <class T>
Array2D<T>::Array2D(int m, int n) 
		: data_(m*n), v_(m), m_(m), n_(n) {
	if (m>0 && n>0)
	{
		T* p = &(data_[0]);
		for (int i=0; i<m; i++)
		{
			v_[i] = p;
			p += n;
		}
	}
}



template <class T>
Array2D<T>::Array2D(int m, int n, const T &val) : data_(m*n), v_(m), 
													m_(m), n_(n) 
{
  if (m>0 && n>0)
  {
	data_ = val;
	T* p  = &(data_[0]);
	for (int i=0; i<m; i++)
	{
			v_[i] = p;
			p += n;
	}
  }
}

template <class T>
Array2D<T>::Array2D(int m, int n, T *a) : data_(m*n, a), v_(m), m_(m), n_(n)
{
  if (m>0 && n>0)
  {
	T* p = &(data_[0]);
	
	for (int i=0; i<m; i++)
	{
			v_[i] = p;
			p += n;
	}
  }
}


template <class T>
inline T* Array2D<T>::operator[](int i) 
{ 
#ifdef TNT_BOUNDS_CHECK
	assert(i >= 0);
	assert(i < m_);
#endif

return v_[i]; 

}


template <class T>
inline const T* Array2D<T>::operator[](int i) const
{ 
#ifdef TNT_BOUNDS_CHECK
	assert(i >= 0);
	assert(i < m_);
#endif

return v_[i]; 

}

template <class T>
Array2D<T> & Array2D<T>::operator=(const T &a)
{
	/* non-optimized, but will work with subarrays in future verions */

	for (int i=0; i<m_; i++)
		for (int j=0; j<n_; j++)
		v_[i][j] = a;
	return *this;
}




template <class T>
Array2D<T> Array2D<T>::copy() const
{
	Array2D A(m_, n_);

	for (int i=0; i<m_; i++)
		for (int j=0; j<n_; j++)
			A[i][j] = v_[i][j];


	return A;
}


template <class T>
Array2D<T> & Array2D<T>::inject( const Array2D &A ){
	// modified by JTB - copy as much as possible if array dimensions are not equal
	int iI = min( m_, A.m_ );
	int iJ = min( n_, A.n_ );
	for( int i=0; i<iI; i++ )
		for( int j=0; j<iJ; j++ )
			v_[i][j] = A[i][j];
	return * this;
	}

template <class T>
Array2D<T> & Array2D<T>::inject( int iRows, int iCols, const T ** pdA ){
	for( int i=0; i<iRows; i++ )
		for( int j=0; j<iCols; j++ )
			v_[i][j] = pdA[i][j];
	return * this;
	}


template <class T>
Array2D<T> & Array2D<T>::ref(const Array2D<T> &A)
{
	if (this != &A)
	{
		v_ = A.v_;
		data_ = A.data_;
		m_ = A.m_;
		n_ = A.n_;
		
	}
	return *this;
}



template <class T>
Array2D<T> & Array2D<T>::operator=(const Array2D<T> &A)
{
	return ref(A);
}

template <class T>
inline int Array2D<T>::dim1() const { return m_; }

template <class T>
inline int Array2D<T>::dim2() const { return n_; }


template <class T>
Array2D<T>::~Array2D() {}




template <class T>
inline Array2D<T>::operator T**()
{
	return &(v_[0]);
}

template <class T>
inline Array2D<T>::operator const T**()
{
	return &(v_[0]);
}


/* ............... extended interface ............... */
/**
	Create a new view to a subarray defined by the boundaries
	[i0][i0] and [i1][j1].  The size of the subarray is
	(i1-i0) by (j1-j0).  If either of these lengths are zero
	or negative, the subarray view is null.

*/
template <class T>
Array2D<T> Array2D<T>::subarray(int i0, int i1, int j0, int j1) 
{
	Array2D<T> A;
	int m = i1-i0+1;
	int n = j1-j0+1;

	/* if either length is zero or negative, this is an invalide
		subarray. return a null view.
	*/
	if (m<1 || n<1)
		return A;

	A.data_ = data_;
	A.m_ = m;
	A.n_ = n;
	A.v_ = Array1D<T*>(m);
	T* p = &(data_[0]) + i0 *  n_ + j0;
	for (int i=0; i<m; i++)
	{
		A.v_[i] = p + i*n_;

	}	
	return A;
}

template <class T>
inline int Array2D<T>::ref_count()
{
	return ref_count_data();
}



template <class T>
inline int Array2D<T>::ref_count_data()
{
	return data_.ref_count();
}

template <class T>
inline int Array2D<T>::ref_count_dim1()
{
	return v_.ref_count();
}




} /* namespace TNT */

#endif
/* TNT_ARRAY2D_H */

